/* tape_clean_1 - Program to read a sampled game tape, and output a clean square-wave version.

Coding history:

WJB	05/11/16 First draft.

*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

typedef unsigned char byte;
typedef unsigned short word;
typedef int BOOLEAN;
#define	TRUE 1
#define	FALSE 0

/* Structures to define a WAV file for reading or writing */

typedef enum { fmtUnk, fmtW08, fmtW16, fmtW32 } TapeFmt;

typedef struct
	{
	const char *	psFile;		/* File name */
	FILE *			pfTape;		/* File stream */
	TapeFmt			fmt;		/* File format */
	word			nChan;		/* Number of channels in WAV file */
	unsigned int	nRate;		/* WAV file sample rate */
	unsigned int	nSamp;		/* Number of samples remaining */
	BOOLEAN			bSave;		/* True if saving to file */
	} WaveDef;

/* Output error message and terminate program */

void fatal (const char *fmt, ...)
	{
	va_list	vars;
	va_start (vars, fmt);
	vfprintf (stderr, fmt, vars);
	va_end (vars);
	fprintf (stderr, "\n");
	exit (1);
	}

/* Open a WAV file for reading and determine format */

void tape_open_wav (WaveDef *pwd)
	{
	unsigned int nLen;
	char sID[5];
	pwd->fmt = fmtUnk;
	pwd->bSave = FALSE;
	pwd->pfTape = fopen (pwd->psFile, "rb");
	if ( pwd->psFile == NULL ) fatal ("Failed to open input file %s", pwd->psFile);
	fread (sID, sizeof (char), 4, pwd->pfTape);
	sID[4] = '\0';
	if ( strcmp (sID, "RIFF") )	fatal ("Tape WAV file does not have a RIFF header");
	fread (&nLen, sizeof (nLen), 1, pwd->pfTape);
	fread (sID, sizeof (char), 4, pwd->pfTape);
	if ( strcmp (sID, "WAVE") ) fatal ("Tape WAV file is not WAVE format");
	while ( fread (sID, sizeof (char), 4, pwd->pfTape) == 4 )
		{
		fread (&nLen, sizeof (nLen), 1, pwd->pfTape);
		if ( ! strcmp (sID, "fmt ") )
			{
			word wFmt;
			unsigned int nBPS;
			word wAlign;
			unsigned int nBits;
			word wBits;
			fread (&wFmt, sizeof (wFmt), 1, pwd->pfTape);
			if ( wFmt != 1 ) fatal ("Tape WAV file is compressed");
			fread (&pwd->nChan, sizeof (pwd->nChan), 1, pwd->pfTape);
			fread (&pwd->nRate, sizeof (pwd->nRate), 1, pwd->pfTape);
			fread (&nBPS, sizeof (nBPS), 1, pwd->pfTape);
			fread (&wAlign, sizeof (wAlign), 1, pwd->pfTape);
			if ( nLen == 16 )
				{
				fread (&wBits, sizeof (wBits), 1, pwd->pfTape);
				nBits = wBits;
				}
			else if ( nLen == 18 )
				{
				fread (&nBits, sizeof (nBits), 1, pwd->pfTape);
				}
			else
				{
				fatal ("Tape WAV file has unexpected fmt chunk length");
				}
			if ( nBits == 8 )       pwd->fmt = fmtW08;
			else if ( nBits == 16 )	pwd->fmt = fmtW16;
			else if ( nBits == 32 ) pwd->fmt = fmtW32;
			else fatal ("Tape WAV file has unsupported sample length");
			}
		else if ( ! strcmp (sID, "data") )
			{
			if ( pwd->fmt == fmtUnk ) fatal ("No fmt chunk in WAV file before data");
			if ( pwd->fmt == fmtW08 )		pwd->nSamp = nLen / pwd->nChan;
			else if ( pwd->fmt == fmtW16 )	pwd->nSamp = nLen / pwd->nChan / 2;
			else if ( pwd->fmt == fmtW32 )	pwd->nSamp = nLen / pwd->nChan / 4;
			return;
			}
		else
			{
			fseek (pwd->pfTape, nLen, SEEK_CUR);
			}
		}
	fatal ("No data chunk found in Tape WAV file");
	}

/* Read a value from a WAV file (averages over all channels if more than one).
   Returns TRUE if a value read, otherwise FALSE */

BOOLEAN tape_read_wav (WaveDef *pwd, float *pfVal)
	{
	float fVal = 0.0;
	int iChan;
	int nRead;
	byte by;
	short wd;
	float f;
	if ( ( pwd == NULL ) || ( pwd->bSave == TRUE ) || ( pwd->nSamp == 0 ) || ( pwd->pfTape == NULL ) )
		return FALSE;
	for ( iChan = 0; iChan < pwd->nChan; ++iChan)
		{
		switch (pwd->fmt)
			{
			case fmtW08:
				nRead = fread (&by, sizeof (by), 1, pwd->pfTape);
				fVal += by / 128.0 - 1.0;
				break;
			case fmtW16:
				nRead = fread (&wd, sizeof (wd), 1, pwd->pfTape);
				fVal += wd / 32768.0;
				break;
			case fmtW32:
				nRead = fread (&f, sizeof (f), 1, pwd->pfTape);
				fVal += f;
				break;
			}
		if ( nRead == 0 )
			{
			pwd->nSamp = 0;
			return FALSE;
			}
		}
	*pfVal = fVal / pwd->nChan;
	--pwd->nSamp;
	return TRUE;
	}

/* Opens a WAV file for output. Assumes single channel, 8-bit */

void tape_save_wav (WaveDef *pwd)
	{
	int iVal;
	word wVal;
	pwd->bSave = TRUE;
	pwd->fmt = fmtW08;
	pwd->nChan = 1;
	if ( pwd->psFile != NULL )
		{
		pwd->pfTape = fopen (pwd->psFile, "wb");
		fwrite ("RIFF", sizeof (char), 4, pwd->pfTape);
		pwd->nSamp = 0;
		fwrite (&pwd->nSamp, sizeof (pwd->nSamp), 1, pwd->pfTape);
		fwrite ("WAVE", sizeof (char), 4, pwd->pfTape);
		fwrite ("fmt ", sizeof (char), 4, pwd->pfTape);
		iVal = 18;	/* Chunk size */
		fwrite (&iVal, sizeof (iVal), 1, pwd->pfTape);
		wVal = 1;	/* Format */
		fwrite (&wVal, sizeof (wVal), 1, pwd->pfTape);
		wVal = 1;	/* Number of channels */
		fwrite (&wVal, sizeof (wVal), 1, pwd->pfTape);
		fwrite (&pwd->nRate, sizeof (pwd->nRate), 1, pwd->pfTape);
		fwrite (&pwd->nRate, sizeof (pwd->nRate), 1, pwd->pfTape);
		wVal = 1;	/* Block align */
		fwrite (&wVal, sizeof (wVal), 1, pwd->pfTape);
		iVal = 8;	/* Bits per sample */
		fwrite (&iVal, sizeof (iVal), 1, pwd->pfTape);
		fwrite ("data", sizeof (char), 4, pwd->pfTape);
		fwrite (&pwd->nSamp, sizeof (pwd->nSamp), 1, pwd->pfTape);
		}
	else
		{
		pwd->pfTape = NULL;
		}
	}

/* Writes a value to a WAV file */

void tape_write_wav (WaveDef *pwd, byte by)
	{
	if ( ( pwd->pfTape != NULL ) && ( pwd->bSave ) )
		{
		fwrite (&by, sizeof (by), 1, pwd->pfTape);
		++pwd->nSamp;
		}
	}

/* Closes an input or output WAV file. Updates record lengths for output files */

void tape_close_wav (WaveDef *pwd)
	{
	if ( pwd->pfTape != NULL )
		{
		if ( pwd->bSave )
			{
			/* Size of data chunk */
			fseek (pwd->pfTape, 42, SEEK_SET);
			fwrite (&pwd->nSamp, sizeof (pwd->nSamp), 1, pwd->pfTape);
			/* Size of WAVE file */
			pwd->nSamp += 34;
			fseek (pwd->pfTape, 4, SEEK_SET);
			fwrite (&pwd->nSamp, sizeof (pwd->nSamp), 1, pwd->pfTape);
			fseek (pwd->pfTape, 0, SEEK_END);
			}
		fclose (pwd->pfTape);
		pwd->pfTape == NULL;
		}
	}

/* Process a game file */

int main (int nArg, const char **ppsArg)
	{
	const char *psFmt[] = { "Unknown", "8-Bit", "16-Bit", "32-Bit" };
	WaveDef	wdIn;
	WaveDef wdOut;
	double dHys = 0.2;
	double dMean = 0.0;
	double dVar = 0.0;
	double dMin = 1.0;
	double dMax = -1.0;
	float fVal;
	unsigned int nSamp = 0;
	unsigned int nDurn = 0;
	unsigned int nPulse[5] = { 0, 0, 0, 0, 0 };
	byte byOut;
	byte byLast;

	/* Usage message if no parameters */
	
	if ( nArg < 2 )
		{
		fprintf (stderr,
			"Converts a sampled Wave file to a clean square wave.\n"
			"The output file is single channel, 8-bit with the same sample rate as the input.\n\n"
			"Usage: %s [--hys value] input [output]\n\n"
			"\tvalue =\tHysteresis value (0.0-1.0), default 0.2.\n"
			"\tinput =\tInput file name or path.\n"
			"\toutput =\tOutput file name (if omitted just produce statistics).\n",
			ppsArg[0]);
		exit (1);
		}

	/* Read hysterisis parameter if not default */
	
	if ( ( nArg > 3 ) && ( strcmp (*ppsArg, "--hys") == 0 ) )
		{
		dHys = atof (ppsArg[1]);
		ppsArg +=2;
		nArg -= 2;
		if ( ( dHys < 0.0 ) || ( dHys > 1.0 ) ) fatal ("Hysterisis value must be between 0.0 and 1.0");
		}

	/* Get input and output file names */
	
	wdIn.psFile = ppsArg[1];
	wdOut.psFile = NULL;
	if ( nArg > 2 )	wdOut.psFile = ppsArg[2];

	/* Read input file and collect statistics */
	
	printf ("First pass: Reading %s\n", wdIn.psFile);
	tape_open_wav (&wdIn);
	printf ("\tFormat:\t%s\n"
		"\tChannels:\t%d\n"
		"\tSample Rate:\t%d/s\n",
		psFmt[wdIn.fmt],
		wdIn.nChan,
		wdIn.nRate);
	while ( tape_read_wav (&wdIn, &fVal) )
		{
		++nSamp;
		dMean += fVal;
		dVar += fVal * fVal;
		if ( fVal < dMin ) dMin = fVal;
		if ( fVal > dMax ) dMax = fVal;
		}
	tape_close_wav (&wdIn);
	dMean /= nSamp;
	dVar = dVar / nSamp - dMean * dMean;
	if ( dVar > 0.0 ) dVar = sqrt (dVar);
	printf ("\tMean:\t%6.3f\n"
		"\tStd. Dev:\t%6.3f\n"
		"\tMinimum:\t%6.3f\n"
		"\tMaximum:\t%6.3f\n",
		dMean,
		dVar,
		dMin,
		dMax);
	if ( dVar < 0.0 ) fatal ("No variation in input data");

	/* Produce a clean output file and check pulse durations */
	
	if ( wdOut.psFile != NULL ) printf ("Second pass: writing %s\n", wdOut.psFile);
	else printf ("Second pass\n");
	tape_open_wav (&wdIn);
	wdOut.nRate = wdIn.nRate;
	tape_save_wav (&wdOut);
	tape_read_wav (&wdIn, &fVal);
	fVal = ( fVal - dMean ) / dVar;
	if ( fVal >= 0.0 ) byOut = 255;
	else byOut = 0;
	byLast = byOut;
	tape_write_wav (&wdOut, byOut);
	nDurn = 0;
	while ( tape_read_wav (&wdIn, &fVal) )
		{
		fVal = ( fVal - dMean ) / dVar;
		if ( byOut == 0 )
			{
			if ( fVal > dHys ) byOut = 255;
			}
		else
			{
			if ( fVal < -dHys ) byOut = 0;
			}
		tape_write_wav (&wdOut, byOut);
		++nDurn;
		if ( byOut != byLast )
			{
			nDurn = wdOut.nRate / nDurn;
			if ( nDurn < 2000 )			++nPulse[0];
			else if ( nDurn < 2800 )	++nPulse[1];
			else if ( nDurn < 4000 )	++nPulse[2];
			else if ( nDurn < 5600 )	++nPulse[3];
			else						++nPulse[4];
			nDurn = 0;
			}
		byLast = byOut;
		}
	tape_close_wav (&wdIn);
	tape_close_wav (&wdOut);
	printf ("\tLong pulses:\t%d\n"
		"\tOne pulses:\t%d\n"
		"\tMiddle pulses:\t%d\n"
		"\tZero pulses:\t%d\n"
		"\tShort pulses:\t%d\n",
		nPulse[0],
		nPulse[1],
		nPulse[2],
		nPulse[3],
		nPulse[4]);

	return 0;
	}
